/**
 * Copyright (c) 2022
 *    Marcus Britanicus (https://gitlab.com/marcusbritanicus)
 *    Abrar (https://gitlab.com/s96abrar)
 *
 * This file was originally a part of SmartVerticalFlowLayout project
 * <https://github.com/VaysseB/SmartVerticalFlowLayout>
 * Several modifications have been made to suit the needs of DFL.
 * Any and all bug reports are to be filed with DFL and not SVFL.
 *
 * DFL::DynamicLayout is a dynamic layout that switches from horizontal
 * to vertical layout mode if the available width goes below a
 * particular value (default: 300 px).
 **/

#include <cmath>
#include "DFDynamicLayout.hpp"

DFL::DynamicLayout::DynamicLayout( QWidget *parent ) : QLayout( parent ) {
    setContentsMargins( -1, -1, -1, -1 );
}


DFL::DynamicLayout::~DynamicLayout() {
    clear();
}


int DFL::DynamicLayout::horizontalSpacing() const {
    if ( m_hSpace >= 0 ) {
        return m_hSpace;
    }

    return smartSpacing( QStyle::PM_LayoutHorizontalSpacing );
}


int DFL::DynamicLayout::verticalSpacing() const {
    if ( m_vSpace >= 0 ) {
        return m_vSpace;
    }

    return smartSpacing( QStyle::PM_LayoutVerticalSpacing );
}


void DFL::DynamicLayout::setSpacing( int space ) {
    setHorizontalSpacing( space );
    setVerticalSpacing( space );

    doLayout( geometry(), false );
}


void DFL::DynamicLayout::setHorizontalSpacing( int space ) {
    m_isLayoutModified |= (m_hSpace != space);
    m_hSpace            = space;

    doLayout( geometry(), false );
}


void DFL::DynamicLayout::setVerticalSpacing( int space ) {
    m_isLayoutModified |= (m_vSpace != space);
    m_vSpace            = space;

    doLayout( geometry(), false );
}


bool DFL::DynamicLayout::hasHeightForWidth() const {
    return true;
}


int DFL::DynamicLayout::heightForWidth( int width ) const {
    doLayout( QRect( 0, 0, width, 1 ), true );
    return m_structureGeometry.height();
}


int DFL::DynamicLayout::maxRowCount() const {
    return m_maxRowCount;
}


void DFL::DynamicLayout::setMaxRowCount( int count ) {
    m_isLayoutModified |= (m_maxRowCount != count);
    m_maxRowCount       = count;

    doLayout( geometry(), false );
}


int DFL::DynamicLayout::count() const {
    return m_items.size();
}


void DFL::DynamicLayout::addItem( QLayoutItem *item ) {
    /**
     * m_isLayoutModified = true; is useless here as
     * long as QLayout keeps calling setGeometry
     */
    m_items.append( item );
}


QLayoutItem *DFL::DynamicLayout::itemAt( int index ) const {
    return m_items.value( index );
}


QLayoutItem *DFL::DynamicLayout::takeAt( int index ) {
    m_isLayoutModified = true;
    QLayoutItem *item = m_items.takeAt( index );

    doLayout( geometry(), false );

    return item;
}


void DFL::DynamicLayout::clear() {
    QLayoutItem *item;

    Q_FOREACH ( item, m_items ) {
        delete item;
    }
}


void DFL::DynamicLayout::setAlignment( Qt::Alignment align ) {
    m_isLayoutModified = (alignment() != align);
    QLayout::setAlignment( align );

    doLayout( geometry(), false );
}


Qt::Orientations DFL::DynamicLayout::expandingDirections() const {
    if ( (alignment() == 0) || (alignment() == Qt::AlignJustify) ) {
        return Qt::Horizontal;
    }

    return Qt::Orientations();
}


void DFL::DynamicLayout::setGeometry( const QRect& rect ) {
    m_isLayoutModified |= (rect != geometry() );

    doLayout( rect, false );

    QLayout::setGeometry( rect );
}


QSize DFL::DynamicLayout::sizeHint() const {
    return m_structureGeometry.size();
}


QSize DFL::DynamicLayout::minimumSizeHint() const {
    return m_structureGeometry.size();
}


QSize DFL::DynamicLayout::minimumSize() const {
    return sizeHint();
}


void DFL::DynamicLayout::updateLayout() {
    m_isLayoutModified = true;
    doLayout( geometry(), false );
}


void DFL::DynamicLayout::doLayout( const QRect& rect, bool dummy ) const {
    if ( !m_isLayoutModified ) {
        return;
    }

    int left, top, right, bottom;

    getContentsMargins( &left, &top, &right, &bottom );
    QRect effectiveRect = rect.adjusted( +left, +top, -right, -bottom );
    int   x             = effectiveRect.x();

    // update structure ( ordering line by lines )
    m_structure.clear();
    QLayoutItem          *item = 0;
    QList<QLayoutItem *> rowItems;

    Q_FOREACH ( item, m_items ) {
        QSize itemSize = item->sizeHint();

        // if the item is over the line limit OR limit reach per row
        if ( (x + itemSize.width() > effectiveRect.right() + 1) || ( (m_maxRowCount > 0) && (rowItems.count() >= m_maxRowCount) ) ) {
            // if the line isn't empty => we add to the structure
            // and reset the row
            if ( !rowItems.isEmpty() ) {
                m_structure.append( rowItems );
                x = effectiveRect.x();
            }

            rowItems.clear();
        }

        rowItems.append( item );
        x += itemSize.width() + horizontalSpacing();
    }

    if ( !rowItems.isEmpty() ) {
        m_structure.append( rowItems );
    }

    int y = effectiveRect.y();
    // update items position
    // width and height msut be positive ! ( otherwise, infinite loop )
    QRect layoutGeometry = QRect( 0, 0, 1, 1 );

    Q_FOREACH ( rowItems, m_structure ) {
        // reset horizontal position
        x = effectiveRect.x();

        // calculating row height and row width
        int rowHeight = 0;
        int rowWidth  = 0;
        Q_FOREACH ( item, rowItems ) {
            QSize itemSize = item->sizeHint();
            rowWidth += itemSize.width();
            rowHeight = qMax( rowHeight, itemSize.height() );
        }

        // placing X coordinates
        int hSpaceTot = horizontalSpacing() * (rowItems.count() - 1);
        int freeSpace = effectiveRect.width() - rowWidth - hSpaceTot;

        // if stretch content => nothing to do ( free space is ignored )
        // if no strect and center
        if ( (alignment() & Qt::AlignHCenter) ) {
            // shift to align center
            x += freeSpace / 2;
        }
        // if no strect and left
        else if ( (alignment() & Qt::AlignLeft) ) {
            // no shift X ( already align to the left )
        }
        // if no strect and right
        else if ( (alignment() & Qt::AlignRight) ) {
            // shift to align the right
            x += freeSpace;
        }

        qreal cumulStrecthWidth = x;

        // setting position of row elements
        Q_FOREACH ( item, rowItems ) {
            QSize itemSize = item->sizeHint();

            QRect itemGeometry;

            // if no strect and align
            if ( (alignment() & Qt::AlignHCenter) ||
                 (alignment() & Qt::AlignLeft) ||
                 (alignment() & Qt::AlignRight) ) {
                itemGeometry = QRect( x, y, itemSize.width(), rowHeight );

                // moving to next element position
                x += itemSize.width() + horizontalSpacing();
            }
            // if stretch content
            else {
                // getting true width depending on the pourcentage of width taken by the item in the row
                qreal strecthOptimalWidth = (effectiveRect.width() - hSpaceTot) *
                                            ( ( qreal )itemSize.width() ) / ( ( qreal )rowWidth);

                itemGeometry = QRect( x, y, std::floor( strecthOptimalWidth ), rowHeight );

                qreal xIncr = strecthOptimalWidth + horizontalSpacing();
                cumulStrecthWidth += xIncr;
                x = std::floor( cumulStrecthWidth );
            }

            // update geometry of item
            if ( not dummy ) {
                item->setGeometry( itemGeometry );
            }

            // update internal geometry
            layoutGeometry = layoutGeometry.united( itemGeometry );
        }

        y += rowHeight + verticalSpacing();
    }

    layoutGeometry.setHeight( layoutGeometry.height() + bottom );
    layoutGeometry.setWidth( layoutGeometry.width() + right );
    m_structureGeometry = layoutGeometry;

    m_isLayoutModified = false;
}


int DFL::DynamicLayout::smartSpacing( QStyle::PixelMetric pm ) const {
    QObject *parent       = this->parent();
    QLayout *parentLayout = 0;

    int space = 0;

    if ( parent && parent->isWidgetType() ) {
        QWidget *pw = static_cast<QWidget *>(parent);
        space = pw->style()->pixelMetric( pm, 0, pw );
    }

    else if ( (parentLayout = qobject_cast<QLayout *>( parent ) ) ) {
        space = parentLayout->spacing();
    }

    return qMax( space, 0 );
}
